pypy generate_unicodedb.py --unidata_version=3.2.0 --output=unicodedb_3_2_0
pypy generate_unicodedb.py --unidata_version=5.2.0 --output=unicodedb_5_2_0 --base=unicodedb_3_2_0
pypy generate_unicodedb.py --unidata_version=9.0.0 --output=unicodedb_9_0_0 --base=unicodedb_3_2_0
pypy generate_unicodedb.py --unidata_version=11.0.0 --output=unicodedb_11_0_0 --base=unicodedb_3_2_0 
pypy generate_unicodedb.py --unidata_version=12.1.0 --output=unicodedb_12_1_0 --base=unicodedb_3_2_0 
pypy generate_unicodedb.py --unidata_version=13.0.0 --output=unicodedb_13_0_0 --base=unicodedb_3_2_0 
pypy generate_unicodedb.py --unidata_version=14.0.0 --output=unicodedb_14_0_0 --base=unicodedb_3_2_0
