
import py; py.test.skip("not implemented: get_all_loop_runs()")

from rpython.jit.backend.llsupport.test.ztranslation_test import TranslationTestJITStats


class TestTranslationJITStatsPPC(TranslationTestJITStats):
    pass
