
from rpython.jit.backend.zarch.test.support import JitZARCHMixin
from rpython.jit.backend.test.jitlog_test import LoggerTest

class TestJitlog(JitZARCHMixin, LoggerTest):
    # for the individual tests see
    # ====> ../../../test/jitlog_test.py
    pass
